import datetime, requests, urlparse, collections
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import statsmodels.api as sm
from scipy import interpolate
from matplotlib.collections import LineCollection
from sklearn.linear_model import LinearRegression
from sklearn.isotonic import IsotonicRegression


# GLOBAL VARIABLE
BASE_URL = 'https://proba-v-mep.esa.int/api/timeseries/v1.0/ts/'

class TimeSeries(object):
    def __init__(self, coverageId, lon, lat, start_date, end_date):
        self.coverageId = coverageId
        self.start_date = start_date
        self.end_date = end_date
        self.x = lat
        self.y = lon

        tsvURL = urlparse.urljoin(BASE_URL, self.coverageId + '/')
        tsvURL = urlparse.urljoin(tsvURL, 'point')

        payload = {
            'lon': str(self.y),
            'lat': str(self.x)
        }

        if self.start_date != None:
            payload['startDate'] = self.start_date

        if self.end_date != None:
            payload['endDate'] = self.end_date

        response = requests.get(tsvURL, params=payload)
        _timeSeries = response.json()['results']

        self.data = pd.DataFrame(
            map(lambda x: [datetime.datetime.strptime(x['date'], "%Y-%m-%d").date(), x['result']['average']],
                _timeSeries))
        self.data = self.data.rename(index=str, columns={0: 'DateTime', 1: 'data'})
        self.data = self.data.set_index('DateTime')
        self.data.index = pd.DatetimeIndex(self.data.index)

def isotonic_vs_linear_regression(series):
    n = len(series)
    x = np.arange(n)
    y = series.values

    # Fit IsotonicRegression and LinearRegression models
    ir = IsotonicRegression()

    y_ = ir.fit_transform(x, y)

    lr = LinearRegression()
    lr.fit(x[:, np.newaxis], y)

    segments = [[[i, y[i]], [i, y_[i]]] for i in range(n)]
    lc = LineCollection(segments, zorder=0)
    lc.set_array(np.ones(len(y)))
    lc.set_linewidths(0.5 * np.ones(n))

    fig = plt.figure()
    plt.plot(x, y, 'r.', markersize=12)
    plt.plot(x, y_, 'g.-', markersize=12)
    plt.plot(x, lr.predict(x[:, np.newaxis]), 'b-')
    plt.gca().add_collection(lc)
    plt.legend(('Data', 'Isotonic Fit', 'Linear Fit'), loc='lower right')
    plt.title('Isotonic regression')
    plt.show()

# Get Data
lon, lat = -89.64363, 18.96943
ndvi = TimeSeries('PROBAV_L3_S10_TOC_NDVI_333M', lon, lat, '2013-01-01', '2018-01-15')
rainfall = TimeSeries('CHIRPS_RAINFALL', lon, lat, '2013-01-01', '2018-01-15')

r = isotonic_regression(ndvi.data)
print(r)