'''
    author: rkopeinig
    script: Advanced Multivariate Analysis
    description: Bunch of algorithms witihn advanced multivariate analysis.
                 See Proba-V Time-Series Analysis Toolbox Poster.
    version: 0.2
'''

import numpy as np
import pandas as pd

import statsmodels.api as sm
from sklearn.isotonic import IsotonicRegression
from sklearn.linear_model import LinearRegression
from sklearn.decomposition import PCA
from statsmodels.tsa.arima_model import ARIMA


def arima_model(series):
    """ Autoregressive Integrated Moving Average ARIMA Model.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return param RSS: RSS of ARIMA result..
        :rtype: float

        :return log_series_diff: Pandas DataFrame of log series difference.
        :rtype: pandas.core.series.Series

        :return result_arima: ARIMA result.
        :rtype: statsmodels.tsa.arima_model.ARIMAResultsWrapper

    """
    log_series = series.apply(lambda x: np.log(x))
    log_series_diff = log_series - log_series.shift()

    model = ARIMA(log_series, order=(1, 1, 1))
    results_ARIMA = model.fit(disp=-1)

    RSS = np.nansum((results_ARIMA.fittedvalues - log_series_diff) ** 2)
    return RSS, log_series_diff, results_ARIMA


def auto_regressive_model(series):
    """ Autoregressive Model.

        :param series: First One-dimensional ndarray with axis labels including Time-Series.
        :type series: pandas.core.series.Series

        :return param RSS: RSS of AR result
        :rtype: float

        :return log_series_diff: Pandas DataFrame of log series difference.
        :rtype: pandas.core.series.Series

        :return result_ar: Autoregressive Model result.
        :rtype: statsmodels.tsa.arima_model.ARIMAResultsWrapper

    """
    log_series = series.apply(lambda x: np.log(x))
    log_series_diff = log_series - log_series.shift()

    model = ARIMA(log_series, order=(2, 1, 0))
    results_ar = model.fit(disp=-1)

    RSS = np.nansum((results_ar.fittedvalues - log_series_diff) ** 2)
    return RSS, log_series_diff, results_ar


def moving_average_model(series):
    """ Moving Average.

            :param series: First One-dimensional ndarray with axis labels including Time-Series.
            :type series: pandas.core.series.Series

            :return param RSS: RSS of MA result
            :rtype: float

            :return log_series_diff: Pandas DataFrame of log series difference.
            :rtype: pandas.core.series.Series

            :return result_ma: Moving Average result.
            :rtype: statsmodels.tsa.arima_model.ARIMAResultsWrapper

        """
    log_series = series.apply(lambda x: np.log(x))
    log_series_diff=log_series-log_series.shift()

    model=ARIMA(log_series, order=(0,1,2))
    results_ma = model.fit(disp=-1)

    RSS = np.nansum((results_ma.fittedvalues - log_series_diff) ** 2)
    return RSS, log_series_diff, results_ma


def general_least_squares(series_1, series_2):
    """ Generalized least squares model with a general covariance structures

            :param series_1: One-dimensional ndarray with axis labels including Time-Series.
            :type series_1: pandas.core.series.Series

            :param series_2: One-dimensional ndarray with axis labels including Time-Series.
            :type series_2: pandas.core.series.Series

            :return: Generalized least squares regression result summary
            :rtype: statsmodels.iolib.summary.Summary

    """
    ols_resid = sm.OLS(series_1.values, series_2.values).fit().resid
    resid_fit = sm.OLS(ols_resid[1:], sm.add_constant(ols_resid[:-1])).fit()
    rho = resid_fit.params[1]

    from scipy.linalg import toeplitz
    toeplitz(range(5))
    order = toeplitz(range(len(ols_resid)))

    sigma = rho**order
    gls_model = sm.GLS(series_1.values, series_2.values, sigma=sigma)
    gls_results = gls_model.fit()
    return gls_results.summary()


def isotonic_regression(series):
    """ Isotonic Regression finds a non-decreasing approximation of a function while minimizing the mean squared error on the training data.

            :param series: One-dimensional ndarray with axis labels including Time-Series.
            :type series: pandas.core.series.Series

            :return: Pandas Series of linear isotonic results
            :rtype: pandas.core.series.Series

    """
    x = np.arange(len(series))
    y = series.values
    ir = IsotonicRegression()
    return pd.Series(ir.fit_transform(x,y),index=series.index)


def linear_regression(series):
    """ Ordinary least squares Linear Regression

            :param series: One-dimensional ndarray with axis labels including Time-Series.
            :type series: pandas.core.series.Series

            :return: Pandas Series of linear regression results
            :rtype: pandas.core.series.Series

    """
    x = np.arange(len(series))
    y = series.values
    lr = LinearRegression()
    lr.fit(x[:, np.newaxis], y)
    return pd.Series(lr.predict(x[:, np.newaxis]),index=series.index)

def principal_component_analysis(df, n_components=None):
    """ Principal Component Analysis

        :param df: Two-dimensional size-mutable ndarray with axis labels including Time-Series.
        :type df: pandas.core.series.Series

        :param n_components: Number of components to keep
        :type n_components: int, float, None or string

        :return: Pandas Series of Principal Component Analysis for given Time-Series
        :rtype: pandas.core.series.Series

    """
    pca = PCA(n_components=None)
    pca.fit(df)
    pca = pd.DataFrame(pca.transform(df), columns=['PCA'], index=df.index)
    return(pca['PCA'])